<?php



namespace WpieApp\Core\Helpers;

use WP_Filesystem_Direct;

defined( 'WPIE__FILE__' ) || exit;


trait Filesystem {

        
        private static $fileSystem;

        
        public static function pathJoin( $base = "", $path = "" ) {
                if ( \path_is_absolute( $path ) ) {
                        return $base;
                }

                return rtrim( $base, '/.' ) . DIRECTORY_SEPARATOR . ltrim( $path, '/.' );
        }

        
        public static function adminUrl( $page = '', $args = [] ) {

                $args = \wp_parse_args( $args, [ 'page' => $page ] );

                return \add_query_arg( $args, \admin_url( 'admin.php' ) );
        }

        
        public static function directoryExists( $path = "" ) {

                if ( !Helper::isNonEmptyStr( $path ) ) {
                        return false;
                }

                return is_dir( $path ) && file_exists( $path );
        }

        
        public static function createDir( $path = "" ) {

                if ( self::directoryExists( $path ) ) {
                        return true;
                }

                if ( !self::hasDirPermission( $path ) ) {
                        return false;
                }

                return \wp_mkdir_p( $path );
        }

        
        public static function removeDir( $path = "", $recursive = true ) {

                if ( !self::hasDirPermission( $path ) ) {
                        return false;
                }

                return Helper::getFilesystem()->rmdir( $path, $recursive );
        }

        
        public static function copyDir( $source, $destination ) {

                if ( !self::hasDirPermission( $source ) || !self::hasDirPermission( $destination ) ) {
                        return false;
                }

                // Get the directory listing for the source directory
                $dir_list = Helper::getFilesystem()->dirlist( $source );

                if ( false === $dir_list ) {
                        return false;
                }

                // Recursively copy each file and subdirectory in the source directory
                foreach ( $dir_list as $filename => $fileinfo ) {

                        $src_file = trailingslashit( $source ) . $filename;

                        $dest_file = trailingslashit( $destination ) . $filename;

                        if ( $fileinfo[ 'type' ] === 'f' ) {
                                // If the file is a regular file, copy it
                                if ( !self::duplicate( $src_file, $dest_file, true ) ) {
                                        return false;
                                }
                        } elseif ( $fileinfo[ 'type' ] === "d" ) {
                                // If the file is a directory, create the directory on the destination and recursively copy its contents
                                if ( !self::createDir( $dest_file ) ) {
                                        return false;
                                }
                                if ( !self::copyDir( $src_file, $dest_file ) ) {
                                        return false;
                                }
                        }
                }

                return true;
        }

        
        public static function duplicate( $source, $destination, $overwrite = true, $mode = false ) {

                if ( !self::hasDirPermission( $destination, false ) ) {
                        return false;
                }

                if ( !self::getFilesystem()->exists( $source ) ) {
                        return false;
                }

                return self::getFilesystem()->copy( $source, $destination, $overwrite, $mode );
        }

        
        public static function hasDirPermission( $path = "", $isPluginDir = true ) {

                $dir = $isPluginDir ? WPIE_UPLOAD_DIR : WPIE_UPLOAD_BASE_DIR;

                if ( !Helper::isNonEmptyStr( $path ) || !Helper::strStartsWith( $dir, $path ) || trim( $dir, DIRECTORY_SEPARATOR ) === trim( $path, DIRECTORY_SEPARATOR ) ) {
                        return false;
                }

                return true;
        }

        
        public static function getFilesystem() {

                if ( !self::$fileSystem ) {
                        self::setFilesystem();
                }

                return self::$fileSystem;
        }

        
        private static function setFilesystem() {

                if ( self::$fileSystem ) {
                        return;
                }

                if ( (!class_exists( '\WP_Filesystem_Direct' ) ) && is_readable( ABSPATH . 'wp-admin/includes/class-wp-filesystem-direct.php' ) ) {
                        require_once ABSPATH . 'wp-admin/includes/class-wp-filesystem-base.php';
                        require_once ABSPATH . 'wp-admin/includes/class-wp-filesystem-direct.php';
                }

                self::$fileSystem = new WP_Filesystem_Direct( new \StdClass() );

                /*
                 * Define the timeouts for the connections. Only available after the constructor is called
                 * to allow for per-transport overriding of the default.
                 */
                if ( !defined( 'FS_CONNECT_TIMEOUT' ) ) {
                        define( 'FS_CONNECT_TIMEOUT', 30 );
                }
                if ( !defined( 'FS_TIMEOUT' ) ) {
                        define( 'FS_TIMEOUT', 30 );
                }

                // Set the permission constants if not already set.
                if ( !defined( 'FS_CHMOD_DIR' ) ) {
                        define( 'FS_CHMOD_DIR', ( fileperms( ABSPATH ) & 0777 | 0755 ) );
                }
                if ( !defined( 'FS_CHMOD_FILE' ) ) {
                        define( 'FS_CHMOD_FILE', ( fileperms( ABSPATH . 'index.php' ) & 0777 | 0644 ) );
                }
        }

        
        public static function setWriteble( $dir = "", $mode = 0777 ) {

                if ( !self::hasDirPermission( $dir ) ) {
                        return false;
                }

                if ( is_dir( $dir ) && !file_exists( $dir ) ) {
                        return false;
                }

                if ( \wp_is_writable( $dir ) ) {
                        return true;
                }

                Helper::getFilesystem()->chmod( $dir, $mode );

                if ( \wp_is_writable( $dir ) ) {
                        return true;
                }

                return false;
        }

        
        public static function emptyDir( $dir = "" ) {

                if ( !self::hasDirPermission( $dir ) ) {
                        return false;
                }

                if ( !Helper::getFilesystem()->is_dir( $dir ) ) {
                        return false;
                }

                if ( !self::removeDir( $dir ) ) {
                        return false;
                }

                if ( !self::createDir( $dir ) ) {
                        return false;
                }
                return true;
        }

        
        public static function download( $file = "" ) {

                $allowedFiles = [ WP_CONTENT_DIR . DIRECTORY_SEPARATOR . "debug.log" ];

                if ( !self::hasDirPermission( $file ) && !in_array( $file, $allowedFiles ) ) {
                        Helper::ajaxError( __( 'Access Denied', 'wp-import-export' ) );
                }

                if ( Helper::isEmptyStr( $file ) || !Helper::getFilesystem()->exists( $file ) ) {
                        Helper::ajaxError( __( 'File not Exist', 'wp-import-export' ) );
                }

                header( 'Content-Description: File Transfer' );
                header( 'Content-Type: application/octet-stream' );
                header( 'Content-Disposition: attachment; filename=' . basename( $file ) );
                header( 'Expires: 0' );
                header( 'Cache-Control: must-revalidate' );
                header( 'Pragma: public' );
                header( 'Content-Length: ' . Helper::getFilesystem()->size( $file ) );
                if ( \ob_get_length() > 0 ) {
                        \ob_clean();
                }
                // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
                echo Helper::getFilesystem()->get_contents( $file );
                die();
        }
}
