<?php

namespace WpieApp\Core\Office\Reader\CSV;

use WpieApp\Core\Office\Reader\ReaderInterface;
use WpieApp\Core\Helpers\Helper;
use SplFileObject;

defined( 'WPIE__FILE__' ) || exit;


class Reader {

        
        private $fp;

        
        private $labels = [];

        
        private $data = [];

        
        private $delimiter = ",";

        
        private $enclosure = "'";

        
        private $rowNumber = 0;

        
        private $hasTitle = true;

        
        public function load( $file = "", $options = [] ) {

                $isReadable = Helper::isReadable( $file );

                if ( is_wp_error( $isReadable ) ) {
                        return $isReadable;
                }

                $this->fp = new SplFileObject( $file, 'r' );

                $this->fp->setFlags( SplFileObject::READ_CSV );

                $this->setOptions( $options );

                $this->fp->setCsvControl( $this->delimiter, $this->enclosure );

                $this->rewind();

                $this->setLabels();

                return true;
        }

        
        private function setOptions( $options = [] ) {

                $this->hasTitle = isset( $options[ 'hasTitle' ] ) && is_bool( $options[ 'hasTitle' ] ) ? $options[ 'hasTitle' ] : true;

                //Set Delimiter
                $delimiter = isset( $options[ 'delimiter' ] ) ? $options[ 'delimiter' ] : "";
                $delimiter = Delimiter::get( $this->fp, $delimiter );

                if ( \is_wp_error( $delimiter ) ) {
                        return $delimiter;
                }
                $this->delimiter = \is_string( $delimiter ) && \strlen( $delimiter ) === 1 ? $delimiter : $this->delimiter;

                //Set Delimiter
                $enclosure = isset( $options[ 'enclosure' ] ) ? $options[ 'enclosure' ] : "";
                $enclosure = Enclosure::get( $this->fp, $enclosure, $delimiter );

                if ( \is_wp_error( $enclosure ) ) {
                        return $enclosure;
                }
                $this->enclosure = $enclosure;
        }

        
        public function getLabels() {

                return $this->labels;
        }

        
        private function setLabels() {

                if ( Helper::isNonEmptyArray( $this->labels ) ) {
                        return;
                }

                $this->labels = [];

                $this->next();

                $data = $this->current();

                if ( $this->hasTitle ) {

                        $this->labels = Helper::createUniqueArray( $data );

                        return;
                }

                for ( $i = 1; $i <= count( $data ); $i++ ) {
                        $this->labels[] = "field_" . $i;
                }

                $this->rowNumber = 1;
        }

        
        public function current() {
                return $this->data;
        }

        
        public function get() {

                $this->next();

                return $this->current();
        }

        
        public function key() {
                return $this->rowNumber;
        }

        
        public function nextData() {
                return $this->next();
        }

        
        public function next() {

                if ( !$this->valid() ) {
                        $this->data = [];
                        return false;
                }

                $row = $this->fp->fgetcsv();

                if ( !is_array( $row ) || $this->isEmptyRow( $row ) ) {
                        $this->next();
                        return true;
                }

                $this->rowNumber++;

                if ( $this->rowNumber <= 1 ) {
                        $row = $this->removeBom( $row );
                }

                $row = array_map( [ $this, 'sanitize' ], $row );

                $this->data = (empty( $this->labels ) ? $row : Helper::safeArrayCombine( $this->labels, $row ) );

                return true;
        }

        
        public function goToRow( $offset = 1 ) {

                $Offset = Helper::toInt( $offset );

                if ( $Offset < 1 ) {
                        $this->rewind();
                        $this->next();
                        return;
                }

                if ( !$this->hasTitle ) {
                        $Offset--;
                }

                $this->fp->seek( $Offset );

                $this->next();

                $this->rowNumber = $Offset;
        }

        
        public function count() {

                $this->fp->seek( PHP_INT_MAX );

                $count = $this->fp->key();

                $total = 0;

                for ( $i = $count; $i >= 0; $i-- ) {

                        $this->fp->seek( $i );

                        $row = $this->fp->current();

                        if ( $row !== false && !$this->isEmptyRow( $row ) ) {

                                $total = $i + 1;
                                break;
                        }
                }


                if ( $this->hasTitle && $total > 0 ) {
                        $total--;
                }

                return $total;
        }

        
        public function rewind() {

                if ( $this->rowNumber < 1 ) {
                        return;
                }

                $this->fp->rewind();

                if ( $this->hasTitle ) {

                        $this->next();
                }

                $this->rowNumber = 1;
        }

        
        public function valid() {
                return $this->fp->valid();
        }

        
        private function removeBom( $row = [] ) {

                if ( $this->isEmptyRow( $row ) || !isset( $row[ 0 ] ) ) {
                        return $row;
                }

                $row[ 0 ] = Helper::removeBom( $row[ 0 ] );

                return $row;
        }

        
        private function isEmptyRow( $row = null ) {

                if ( !Helper::isNonEmptyArray( $row ) ) {
                        return false;
                }
                return strlen( trim( implode( "", $row ) ) ) === 0;
        }

        
        private function unescapeSpecialChar( $data = "" ) {

                if ( Helper::isEmptyStr( $data ) ) {
                        return $data;
                }

                $active_content_triggers = [ "'=", "'+", "'-", "'@" ];

                if ( in_array( mb_substr( trim( $data ), 0, 2 ), $active_content_triggers, true ) ) {
                        $data = mb_substr( $data, 1 );
                }
                return $data;
        }

        
        private function sanitize( $data = "" ) {

                if ( Helper::isEmptyStr( $data ) ) {
                        return $data;
                }

                $data = $this->unescapeSpecialChar( $data );

                return $data;
        }

        
        public function getDelimiter() {
                return $this->delimiter;
        }

        
        public function getEnclosure() {
                return $this->enclosure;
        }

        
        public function getPointer() {
                return $this->fp->ftell();
        }

        
        public function setPointer( $poniter = 0, $row = 0 ) {

                fseek( $this->fp, $poniter );

                $this->rowNumber = $row;
        }

        
        public function getRowNumber() {

                return $this->rowNumber;
        }
}
