<?php



namespace WpieApp\Core\Utils;

use DateTime;
use DateTimeZone;
use WpieApp\Core\Helpers\Helper;

defined( 'WPIE__FILE__' ) || exit;


class Date {

        
        private static $time_zone;

        
        const FORMAT = "Y-m-d H:i:s";

        
        const DATEFORMAT = "Y-m-d";
        
        const TIMEFORMAT = "H:i:s";

        
        const SEPARATORS = [ '/', '-', '.', ',' ];

        
        private static function getTimezoneString() {

                if ( function_exists( '\wp_timezone_string' ) ) {
                        return \wp_timezone_string();
                }

                $timezone_string = \get_option( 'timezone_string' );

                if ( $timezone_string ) {
                        return $timezone_string;
                }

                $offset  = ( float ) \get_option( 'gmt_offset' );
                $hours   = ( int ) $offset;
                $minutes = ( $offset - $hours );

                $sign      = ( $offset < 0 ) ? '-' : '+';
                $abs_hour  = abs( $hours );
                $abs_mins  = abs( $minutes * 60 );
                $tz_offset = sprintf( '%s%02d:%02d', $sign, $abs_hour, $abs_mins );

                return $tz_offset;
        }

        
        public static function getTimeZone() {

                if ( !is_null( self::$time_zone ) ) {
                        return self::$time_zone;
                }

                if ( function_exists( '\wp_timezone' ) ) {
                        self::$time_zone = \wp_timezone();
                }

                if ( is_null( self::$time_zone ) ) {

                        $timezone = self::getTimezoneString();

                        if ( empty( $timezone ) ) {
                                $timezone = 'UTC';
                        }

                        self::$time_zone = new DateTimeZone( $timezone );
                }

                return self::$time_zone;
        }

        
        public static function now() {

                return \current_time( 'mysql', true );
        }

        
        public static function getLocalByTimestamp( $timestamp = "", $format = self::FORMAT ) {

                if ( !Helper::isNonEmptyStr( $timestamp ) ) {
                        return "";
                }

                $format = Helper::isNonEmptyStr( $format ) ? $format : self::FORMAT;

                $date = self::getCoreDate( $format, $timestamp );

                return self::getLocal( $date, $format );
        }

        
        public static function getLocal( $date = "", $format = self::FORMAT ) {

                if ( !Helper::isNonEmptyStr( $date ) ) {
                        return "";
                }

                $format = Helper::isNonEmptyStr( $format ) ? $format : self::FORMAT;

                // Convert the DateTime object to the local time using date_i18n()
                return \get_date_from_gmt( $date, $format );
        }

        
        public static function get( $date = "", $format = "" ) {

                $default = self::getCoreDate( self::FORMAT );

                if ( Helper::isEmptyStr( $date ) && Helper::isEmptyStr( $format ) ) {
                        return $default;
                }

                if ( Helper::isEmptyStr( $date ) ) {
                        $date = $default;
                }

                $format = Helper::isEmptyStr( $format ) ? self::FORMAT : $format;

                if ( self::isTimestamp( $date ) ) {
                        return self::getCoreDate( $format, $date );
                }

                if ( !strtotime( $date ) ) {

                        $newDate = self::autoCorrect( $date );

                        if ( $newDate === false ) {
                                $date = $default;
                        } else {
                                $date = $newDate;
                        }
                }

                return self::getCoreDate( $format, strtotime( $date ) );
        }

        public static function isTimestamp( $timestamp = null ) {

                if ( !Helper::isNonEmptyStr( $timestamp ) || !is_numeric( $timestamp ) ) {
                        return false;
                }

                $date = self::getCoreDate( 'd-m-Y H:i:s', $timestamp );

                return $date !== false && strtotime( $date ) === ( int ) $timestamp;
        }

        
        public static function getCoreDate( $format = "", $timestamp = null ) {

                //The date will be converted to the site's local timezone
                //phpcs:ignore WordPress.DateTime.RestrictedFunctions.date_date
                return date( $format, $timestamp );
        }

        
        public static function detectSeparator( $date = "" ) {

                if ( Helper::isEmptyStr( $date ) ) {
                        return false;
                }

                $separator = null;

                foreach ( self::SEPARATORS as $sep ) {

                        if ( Helper::contains( $date, $sep ) ) {
                                $separator = $sep;
                                break;
                        }
                }

                return $separator;
        }

        
        public static function autoCorrect( $date = "" ) {

                if ( Helper::isEmptyStr( $date ) ) {
                        return false;
                }

                $separator = self::detectSeparator( $date );

                if ( Helper::isEmptyStr( $separator ) ) {
                        return false;
                }

                $separators = array_diff( self::SEPARATORS, [ $separator ] );

                $newDate = "";

                foreach ( $separators as $newSep ) {

                        $data = str_replace( $separator, $newSep, $date );

                        if ( strtotime( $data ) ) {
                                $newDate = $data;
                                break;
                        }
                }
                return Helper::isNonEmptyStr( $newDate ) ? $newDate : false;
        }
}
